import io
import multiprocessing
from typing import Generator, Any
import numpy as np
from PIL import Image

from ..image_utils import open_image_from_bytes


def slice_rgb_bit_planes(image_bytes: bytes) -> Generator[tuple[str, bytes], Any, None]:
    update_logger = multiprocessing.get_logger()

    image = open_image_from_bytes(image_bytes)

    update_logger.info("Slicing image into separate bitplanes...")
    for channel in image.getbands():
        c_image = np.array(image.getchannel(channel))
        update_logger.info(f"Channel: {channel}")

        for bit_index in range(0, 8):
            update_logger.info(f"Bit index: {bit_index}")
            slice_name = f"Bitplane {channel}{str(7 - bit_index)}.png"
            bitplane_slice = Image.new("1", image.size)
            chk_val = int(f"0b{bit_index * "0"}1{((7 - bit_index) * "0")}", 2)

            for y in range(c_image.shape[0]):
                for x in range(c_image.shape[1]):
                    src_val = c_image[y][x]
                    if chk_val & src_val > 0:
                        match channel:
                            case "R":
                                bitplane_slice.putpixel((x, y), 1)
                            case "G":
                                bitplane_slice.putpixel((x, y), 1)
                            case "B":
                                bitplane_slice.putpixel((x, y), 1)

            image_bytes_io = io.BytesIO()
            bitplane_slice.save(image_bytes, format="PNG")
            yield slice_name, image_bytes_io.getvalue()



import itertools
import multiprocessing
from typing import Generator, Callable
import numpy as np

from ..errors import BPCSError


def xor_lists(a: list[bool], b: list[bool]) -> list[bool]:
    """
    Given two lists a and b, xor every Nth element in each list and put the results in a list.
    :param a: list #1
    :param b: list #2
    :return: a list containing the XOR-ed values of the two lists
    :raises BPCSError: if given two lists of different length
    """
    if len(a) != len(b):
        raise BPCSError(f"Length of given lists to XOR does not match. {len(a)} != {len(b)}")
    return [x ^ y for x, y in zip(a, b)]


def map_2d_array(arr: np.ndarray, func: Callable) -> np.ndarray:
    """
    Given a 2d numpy array, and a function func, map over each and every element of arr while reassigning it to
    func(arr[i, j]), and returning the resulting 2d numpy array.
    :param arr: 2d numpy array
    :param func: a function
    """
    for i in range(arr.shape[0]):
        for j in range(arr.shape[1]):
            arr[i, j] = func(arr[i, j])
    return arr


def pbc_to_cgc(pixels: np.ndarray) -> np.ndarray:
    """
    Converts pixels from PBC to CGC, the process for this is detailed as such:
    The first bit allways matches, for the rest of the bits we do g[i] = b[i] XOR b[i-1]
    :param pixels: an array describing the pixels that uses PBC to represent values
    :return: an array of pixels that uses CGC to represent values instead of PBC
    """
    update_logger = multiprocessing.get_logger()
    update_logger.info("Converting values from PBC to CGC...")

    def pbc_to_cgc_mapper(planes):
        new_planes = []
        for i in range(planes.shape[1]):
            if i == 0:
                new_planes.append(planes[:, i].tolist())
            else:
                new_planes.append(xor_lists(planes[:, i], planes[:, i - 1]))
        return np.array(new_planes).transpose()

    return map_2d_array(pixels, pbc_to_cgc_mapper)


def cgc_to_pbc(arr: np.ndarray) -> np.ndarray:
    """
    Converts pixels from CGC to PBC, the process for this is detailed as such:
    The first bit allways matches, for the rest of the bits we do b[i] = g[i] XOR b[i-1]
    :param arr: n array of pixels that uses CGC to represent values
    :return: an array of pixels that uses PBC to represent values instead of CGC
    """
    update_logger = multiprocessing.get_logger()
    update_logger.info("Converting values from CGC to PBC...")

    def cgc_to_pbc_mapper(planes):
        new_planes = []
        for i in range(planes.shape[1]):
            if i == 0:
                new_planes.append(planes[:, i].tolist())
            else:
                new_planes.append(xor_lists(planes[:, i], new_planes[i - 1]))
        return np.array(new_planes).transpose()

    return map_2d_array(arr, cgc_to_pbc_mapper)


def bit_list_to_decimal(bits: list[bool]) -> int:
    """
    Convert a given list of singular bits to a decimal value.
    :param bits: the list of bits
    :return: the converted decimal value
    """
    return int(''.join(["1" if x else "0" for x in bits]), 2)


def decimal_to_bit_list(val: int, bit_num: int) -> list[bool]:
    """
    Convert a given decimal value to a list of bits with a given length.
    :param val: the decimal value
    :param bit_num: the number of bits that the bit list should contain
    :return: the converted list of bits
    """
    return [bool(i) for i in [int(x) for x in bin(val)[2:].zfill(bit_num)[:bit_num]]]


class BitPlane:
    """
    The class that manages conversion from Pure Binary Code to Canonical Gray Code and back.
    """
    def __init__(self, pixel_array: np.ndarray, gray=True):
        """
        Initializes a BitPlane object.
        :param pixel_array: an array that describes the pixels of an image
        :param gray: should the array be read in CGC instead of PBC.
        """
        self.arr = pixel_array
        self.gray = gray

    def slice(self) -> np.ndarray:
        """
        Converts the values in self.arr into binary, so that every element in i,j now contains a bit-list that
        describes the decimal value instead the decimal value itself; every binary value has bit_num bits. Also,
        converts the self.arr into CGC if self.gray is True.
        Keeps the shape of self.arr.
        :return: the converted numpy ndarray
        """
        update_logger = multiprocessing.get_logger()
        update_logger.info("Slicing image into bit plane blocks...")

        # reshape to a 1d list binary bit vals
        base_array = np.array([decimal_to_bit_list(i, 8) for i in self.arr.reshape(-1)])

        # reshape back to the same shape as the og array with an added dimension for the bit list
        temp_array = np.reshape(base_array, self.arr.shape + (8,))
        if self.gray:
            temp_array = pbc_to_cgc(temp_array)
        return temp_array

    def stack(self) -> np.ndarray:
        """
        Does the reverse of slicing; converts self.arr to decimal values, so that every element in i,j now contains
        a decimal value instead of a bit-list. Also, converts self.arr back to PBC if self.gray is True. Keeps the
        shape of self.arr.
        """
        update_logger = multiprocessing.get_logger()
        temp_array = self.arr
        if self.gray:
            temp_array = cgc_to_pbc(temp_array)

        def iterate_all_but_last_dim(arr: np.ndarray) -> Generator:
            all_indices = [range(dim_size) for dim_size in arr.shape]
            for ind in itertools.product(*all_indices[:-1]):
                yield ind

        update_logger.info("Stacking bit plane blocks into an image...")
        temp_array = np.reshape([bit_list_to_decimal(temp_array[ind].tolist()) for ind in iterate_all_but_last_dim(temp_array)],
                                temp_array.shape[:-1])
        return temp_array



import math
import multiprocessing
import random
import struct
from enum import IntEnum

from ..errors import TokenError
from .aes_gcm import encrypt, decrypt
from .reed_solomon import pad, unpad
from .seeded_shuffler import shuffle_bytes, unshuffle_bytes

"""
Every parameter but the key has a constant length due to its' algorithms restraints:
- algorithm: 1 byte
- ecc_block_size: 1 byte (0 <= ecc_block_size <= 255).
- ecc_symbol_num: 1 byte (0 < ecc_symbol_num < ecc_block_size).
- verification_tag: 16 bytes (hardcoded into aes encryption algorithm).
- nonce: 16 bytes (hardcoded into aes encryption algorithm).
- update_header: 8 bytes (hardcoded into aes encryption algorithm).
- shuffling seed: 64 bytes (hardcoded into shuffling algorithm).
- key: this will be concatenated to the end of the token so it will be the remaining bytes.

Algorithm specific parameters:
- min_alpha: 8 bytes (per pythons size for floating-point numbers).
- num_of_sacrificed_bits: 1 byte (ranges from 0-8).

From this we can know that the length of any token generated by this algorithm, will be at least the minimum length (
each case has a different length), if its less, then it's not a valid token.
"""


class Algorithms(IntEnum):
    LSB = 1
    BPCS = 2


def generic_wrap(plaintext: bytes, key: bytes, ecc_block_size: int,
                 ecc_symbol_num: int) -> tuple[bytes, bytes, bytes, bytes, bytes]:
    update_logger = multiprocessing.get_logger()

    update_logger.info("Encrypting data...")
    encrypted, verification_tag, nonce, update_header, key = encrypt(plaintext, key)

    update_logger.info("Shuffling data...")
    random.seed()
    seed = random.randbytes(64)
    random.seed(seed)
    shuffled = shuffle_bytes(encrypted, seed)

    update_logger.info("Padding data with error correction...")
    wrapped = pad(shuffled, ecc_block_size, ecc_symbol_num)

    return wrapped, verification_tag, nonce, update_header, seed


def wrap_bpcs(plaintext: bytes, key: bytes, ecc_block_size: int, ecc_symbol_num: int, min_alpha: float):
    """
    Wraps contents. First with encryption, then with error correction.
    :param plaintext: the plaintext to be wrapped
    :param key: the encryption key
    :param ecc_block_size: the error correction block size
    :param ecc_symbol_num: the bytes per block that we want to dedicate for error correction
    :param min_alpha: the minimum complexity coefficient
    :return: the wrapped plaintext, and the matching token
    :raises TokenError: if the created token is invalid
    """
    update_logger = multiprocessing.get_logger()
    update_logger.info("Wrapping data to be embedded with BPCS...")

    algorithm = Algorithms.BPCS.to_bytes(1)
    algorithm = struct.pack("c", algorithm)

    wrapped, verification_tag, nonce, update_header, seed = generic_wrap(plaintext, key, ecc_block_size, ecc_symbol_num)

    update_logger.info("Generating token...")
    token = (algorithm + ecc_block_size.to_bytes(1) + ecc_symbol_num.to_bytes(1) +
             verification_tag + nonce + update_header + seed + struct.pack("d", min_alpha) + key)

    if not len(token) == (115 + len(key)):
        raise TokenError("Token creating parameters are invalid.")

    update_logger.info("Wrapping completed!")
    return bytes(wrapped), token


def wrap_lsb(plaintext: bytes, key: bytes, ecc_block_size: int, ecc_symbol_num: int, num_of_sacrificed_bits: int):
    update_logger = multiprocessing.get_logger()
    update_logger.info("Wrapping data to be embedded with LSB...")

    algorithm = Algorithms.LSB.to_bytes(1)
    algorithm = struct.pack("c", algorithm)

    wrapped, verification_tag, nonce, update_header, seed = generic_wrap(plaintext, key, ecc_block_size, ecc_symbol_num)

    update_logger.info("Generating token...")
    token = (algorithm + ecc_block_size.to_bytes(1) + ecc_symbol_num.to_bytes(1) +
             verification_tag + nonce + update_header + seed +
             struct.pack("c", num_of_sacrificed_bits.to_bytes(1)) + key)

    if not len(token) == (108 + len(key)):
        raise TokenError("Token creating parameters are invalid.")

    update_logger.info("Wrapping completed!")
    return bytes(wrapped), token


def get_base_data_from_token(token: bytes):
    algorithm, token = ord(struct.unpack("c", token[:1])[0]), token[1:]
    ecc_block_size, token = ord(token[:1]), token[1:]
    ecc_symbol_num, token = ord(token[:1]), token[1:]
    verification_tag, token = token[:16], token[16:]
    nonce, token = token[:16], token[16:]
    update_header, token = token[:8], token[8:]
    seed, token = token[:64], token[64:]
    return (ecc_block_size, ecc_symbol_num), (verification_tag, nonce, update_header), seed, token


def get_bpcs_token_info(token: bytes):
    """
    Gets the wrapping and bpcs info from a given token.
    :param token: the token to infer data from
    :return: the token's info in the form:
    (ecc_block_size, ecc_symbol_num), (verification_tag, nonce, update_header, key), min_alpha
    :raises TokenError: if the token is invalid
    """
    update_logger = multiprocessing.get_logger()
    update_logger.info("Extracting data from BPCS token...")
    if not len(token) >= 115:
        raise TokenError("Token length is invalid. Token length must be equal to or greater than 42 bytes.")

    ((ecc_block_size, ecc_symbol_num),
     (verification_tag, nonce, update_header), seed, token) = get_base_data_from_token(token)

    min_alpha, token = struct.unpack("d", token[:8])[0], token[8:]
    key = token
    update_logger.info("Data extraction completed!")
    return (ecc_block_size, ecc_symbol_num), (verification_tag, nonce, update_header, key), seed, min_alpha


def get_lsb_token_info(token: bytes):
    update_logger = multiprocessing.get_logger()
    update_logger.info("Extracting data from LSB token...")

    if not len(token) >= 108:
        raise TokenError("Token length is invalid. Token length must be equal to or greater than 43 bytes.")

    ((ecc_block_size, ecc_symbol_num),
     (verification_tag, nonce, update_header), seed, token) = get_base_data_from_token(token)
    num_of_sacrificed_bits, token = int.from_bytes(struct.unpack("c", token[:1])[0]), token[1:]

    key = token
    update_logger.info("Data extraction completed!")
    return (ecc_block_size, ecc_symbol_num), (verification_tag, nonce, update_header, key), seed, num_of_sacrificed_bits


def unwrap(wrapped: bytes, ecc_block_size: int, ecc_symbol_num: int, verification_tag: bytes, nonce: bytes,
           update_header: bytes, seed: bytes, key: bytes):
    """
    Unwraps content, using the following wrapping parameters.
    :param wrapped: the wrapped content
    :param ecc_block_size: the rs block size
    :param ecc_symbol_num: the number of bytes per block that we want to dedicate for error correction
    :param verification_tag: the GCM tag
    :param nonce: the GCM nonce
    :param update_header: the GCM header
    :param seed: the shuffling seed
    :param key: the encryption key
    :return: the unwrapped content
    :raises ReedSolomonError: if the function was fed invalid parameters, or the data had too many errors for the rs
    algorithm to fix
    :raises ContentWrapperError: if the decryption process fails
    """
    update_logger = multiprocessing.get_logger()
    update_logger.info("Unwrapping extracted data...")

    update_logger.info("Unpadding data...")
    shuffled = unpad(wrapped, ecc_block_size, ecc_symbol_num)

    update_logger.info("Unshuffling data...")
    ciphertext = unshuffle_bytes(shuffled, seed)

    update_logger.info("Decrypting data...")
    dec = decrypt(ciphertext, key, verification_tag, nonce, update_header)

    update_logger.info("Unwrapping completed!")
    return dec


def get_max_unwapped_length(max_bit_embbeding_input_length: int, ecc_block_size: int, ecc_symbol_num: int) -> int:
    return math.floor(
        ((max_bit_embbeding_input_length * (ecc_block_size - ecc_symbol_num))
         /
         (8 * ecc_block_size))
        - 16
    )



import multiprocessing
import numpy as np

from .dimension_computing import compute_all_block_indices
from .block_operations import blocks_to_bits
from ..bit_operations_utils import bits_to_bytes
from ..errors import BPCSExtractError
from .core import calc_bpcs_complexity_coefficient, conjugate
from .initilization_vector import slice_iv_from_accepted_blocks, \
    slice_conj_blocks_from_accepted_blocks


def get_extracting_info_from_accepted_blocks(accepted_blocks: np.ndarray, block_shape: tuple[int, int],
                                             alpha: float) -> tuple[int, int, list[bool], np.ndarray]:
    """
    Extract the extracting info (iv values, conjugation map) from the accepted blocks. This function assumes that the
    relevant blocks start at index = 0 of the block array.
    :param accepted_blocks: the accepted blocks in the vessel image
    :param block_shape: the shape of each bit plane block in the image
    :param alpha: the minimum complexity coefficient of the blocks
    :return: the extracting info with the remaining blocks in the form of: message_block_length, message_remnant_bits_num,
    conjugation_map, remaining_blocks
    """
    (message_block_length, message_remnant_bits_num), remaining_blocks = (
        slice_iv_from_accepted_blocks(accepted_blocks, block_shape, alpha))

    conjugation_map, remaining_blocks = slice_conj_blocks_from_accepted_blocks(remaining_blocks, block_shape, alpha,
                                                                               message_block_length)
    return message_block_length, message_remnant_bits_num, conjugation_map, remaining_blocks


def extract_message_from_remaining_blocks(remaining_blocks: np.ndarray, message_block_length: int,
                                          message_remnant_bits_num: int, conjugation_map: list[bool],
                                          block_shape: tuple[int, int]) -> bytes:
    """
    Extracts the message from the image bit plane blocks, using the data from the iv blocks, and the conjugation map.
    This function assumes that the start of the message blocks is at index = 0 of the remaining blocks
    :param remaining_blocks: the remaining blocks in the image after slicing the iv and conj map blocks
    :param message_block_length: the bit length of the message we want to extract in bits
    :param message_remnant_bits_num: the number of bits that belong to the message in the last message block
    :param conjugation_map: a list of bools describing what message blocks were conjugated
    :param block_shape: the shape of each bit plane block in the image
    :return: the extracted message in bytes
    """
    block_area = block_shape[0] * block_shape[1]
    message_blocks = []
    try:
        for i in range(message_block_length):
            block = remaining_blocks[i]
            if conjugation_map[i]:
                message_blocks.append(conjugate(block))
            else:
                message_blocks.append(block)
    except IndexError:
        # this can only happen when the block length extracted from the image is incorrect, this must be caused by an
        # incorrect minimum complexity coefficient. Which means that the client sent an image-token pair that didn't
        # match
        raise BPCSExtractError("The supplied token and image dont match each other. This might happen when opening the "
                               "token with a text editor.")

    message_bits = blocks_to_bits(np.array(message_blocks))
    message_bits = message_bits[:len(message_bits) - (block_area - message_remnant_bits_num)]

    return bits_to_bytes(message_bits)


def extract_message_from_vessel(vessel_blocks: np.ndarray, alpha: float, block_shape: tuple[int, int]) -> bytes:
    """
    Given an array of vessel blocks in the image, extract the iv and conjugation map blocks, and use them to extract the
    message from the remaining blocks.
    :param vessel_blocks: the images' bit plane blocks
    :param alpha: the minimum complexity coefficient threshold to decide which blocks to accept
    :param block_shape: the shape of each bit plane block in the image
    :return: the extracted message in bytes
    :raises BPCSExtractError: if there is a mismatch between the iv, conj map, and image blocks info
    """
    update_logger = multiprocessing.get_logger()
    update_logger.info("Starting reading process...")
    accepted_blocks = []
    bit_planes = compute_all_block_indices(vessel_blocks.shape, block_shape)

    update_logger.info("Extracting blocks with an appropriate complexity coefficient from image blocks...")
    # get all accepted blocks (every block that has a >= alpha0)
    for bit_plane in bit_planes:
        block = vessel_blocks[tuple(bit_plane)]
        if calc_bpcs_complexity_coefficient(block) >= alpha:
            accepted_blocks.append(block)

    accepted_blocks = np.reshape(accepted_blocks, (len(accepted_blocks),) + block_shape)

    update_logger.info("Extracting initialization vector and conjugation map blocks from accepted blocks...")
    message_block_length, message_remnant_bits_num, conjugation_map, remaining_blocks = (
        get_extracting_info_from_accepted_blocks(accepted_blocks, block_shape, alpha))

    if len(remaining_blocks) < message_block_length:
        raise BPCSExtractError(f"The message block length iv states that the number of message blocks in the image is "
                               f"{message_block_length} but there are only {len(remaining_blocks)} blocks. "
                               f"{message_block_length} < {len(remaining_blocks)}")

    if message_remnant_bits_num > block_shape[0] * block_shape[1]:
        raise BPCSExtractError(f"The message remnant bits number states that the number of bits that belong to the "
                               f"message in a block of total length {block_shape[0] * block_shape[1]} is "
                               f"{message_remnant_bits_num}, this is not possible since the message can only take "
                               f"{block_shape[0] * block_shape[1]} bits per block. Not more.")

    update_logger.info("Using extracing info to extract message from accepted blocks...")
    message = extract_message_from_remaining_blocks(remaining_blocks, message_block_length, message_remnant_bits_num,
                                                    conjugation_map, block_shape)
    update_logger.info("Reading process finished!")
    return message



import io
import multiprocessing
from PIL import Image

from ..image_utils import open_image_from_bytes


def show_diff(img1_bytes: bytes, img2_bytes: bytes, exact_diff: bool) -> tuple[tuple[int, int, int], bytes]:
    update_logger = multiprocessing.get_logger()

    update_logger.info("Loading images...")

    img1 = open_image_from_bytes(img1_bytes)
    img2 = open_image_from_bytes(img2_bytes)

    diff_image = Image.new("RGB", img1.size)

    update_logger.info("Calculating diffrences between the images...")

    r_diff, g_diff, b_diff = 0, 0, 0
    diff_pixels_num = 0
    for w in range(img1.size[0]):
        for h in range(img1.size[1]):
            p1 = img1.getpixel((w, h))
            p2 = img2.getpixel((w, h))

            if p1 != p2:
                diff_pixels_num += 1

                r = abs(p1[0] - p2[0]) # type: ignore
                g = abs(p1[1] - p2[1]) # type: ignore
                b = abs(p1[2] - p2[2]) # type: ignore

                r_diff = max(r, r_diff)
                g_diff = max(g, g_diff)
                b_diff = max(b, b_diff)
                if exact_diff:
                    diff_image.putpixel((w, h), (r, g, b))
                else:
                    diff_image.putpixel((w, h),
                                        tuple(255 if d != 0 else 0 for d in (r, g, b)))
    update_logger.info(f"Found {diff_pixels_num} different pixels")

    image_bytes = io.BytesIO()
    diff_image.save(image_bytes, format="PNG")
    return (r_diff, g_diff, b_diff), image_bytes.getvalue()



import io
import PIL.Image
import numpy as np
from PIL import Image


def open_image_from_bytes(image_bytes: bytes) -> PIL.Image.Image:
    return PIL.Image.open(io.BytesIO(image_bytes)).convert("RGB")


def write_image(out_path: str, image: Image.Image) -> None:
    """
    Saves an image to a given path.
    :param out_path: the given path
    :param image: the image object to be saved
    """
    image.save(out_path, out_path.split('.')[-1])


def image_to_array(im: Image.Image) -> np.ndarray:
    """
    Converts a PIL image to a numpy array
    :param im: the image object to convert
    :return: the numpy array representing the image
    """
    return np.array(im)


def array_to_image(arr: np.ndarray) -> Image.Image:
    """
    Converts a numpy array to a PIL image
    :param arr: the numpy array representing the image
    :return: the converted PIL image
    """
    return Image.fromarray(np.uint8(arr))


def image_to_bytes(image: Image.Image) -> bytes:
    image_bytes = io.BytesIO()
    image.save(image_bytes, format="PNG")
    return image_bytes.getvalue()



import multiprocessing

import numpy as np

from .bit_plane import BitPlane
from .extract import extract_message_from_vessel
from .embed import embed_message_in_vessel
from ..image_utils import open_image_from_bytes, image_to_array, array_to_image, image_to_bytes


class BPCSImage:
    """
    The class that manages the reading, writing, embedding, and extracting data in an PIL image object using BPCS
    steganography.
    """
    def __init__(self, image_bytes: bytes, as_cgc: bool):
        """
        Initializes a new instance of the BPCSImage class.
        :param as_cgc: should the image be read in CGC instead of PBC?
        """
        update_logger = multiprocessing.get_logger()
        self.image_bytes = image_bytes
        self.as_gray = as_cgc
        self.pixels = self.read()
        update_logger.info(f"Loaded image as array with shape {self.pixels.shape}")

    def read(self) -> np.ndarray:
        img = open_image_from_bytes(self.image_bytes)
        pixels = image_to_array(img)
        pixels = BitPlane(pixels, self.as_gray).slice()
        return pixels

    def export(self, pixels: np.ndarray) -> bytes:
        update_logger = multiprocessing.get_logger()
        pixels = BitPlane(pixels, self.as_gray).stack()
        img = array_to_image(pixels)
        update_logger.info("Loaded new bit plane blocks as an image!")
        return image_to_bytes(img)

    def embed(self, message_blocks: np.ndarray, message_bit_length: int, alpha: float,
              check_capacity: bool) -> np.ndarray:
        """
        Embeds the given message blocks into the pixels attribute.
        :param message_blocks: the blocks that describe the message we want to embed
        :param message_bit_length: the length of the message in bits
        :param alpha: the complexity coefficient threshold of the BPCS algorithm
        :param check_capacity: should the program check the images' capacity before starting to embed the message blocks
        :return: the resulting pixels after embedding
        """

        new_arr = np.array(self.pixels, copy=True)
        embedded_arr = embed_message_in_vessel(new_arr, alpha, message_blocks, message_bit_length, (8, 8), check_capacity)
        return embedded_arr

    def extract(self, alpha: float) -> bytes:
        """
        Extracts the message hidden in the pixels attribute.
        :param alpha: the complexity coefficient threshold of the BPCS algorithm
        :return: the extracted message bytes
        """
        return extract_message_from_vessel(self.pixels, alpha, (8, 8))



import multiprocessing
import numpy as np

from .block_operations import bytes_to_blocks
from .core import calc_bpcs_complexity_coefficient, conjugate
from .initilization_vector import build_iv_blocks, build_conjugation_blocks
from ..errors import BPCSError, BPCSCapacityError
from .capacity import calculate_embedding_blocks_num, collect_accepted_blocks


def get_message_blocks_from_bytes(message: bytes) -> tuple[np.ndarray, int]:
    """
    Constructs message blocks from given bytes.
    :param message: the message bytes
    :return: the message blocks derived from the given bytes, and the bit length of the message blocks
    """
    return bytes_to_blocks(message, (8, 8)), len(message) * 8


def get_conjugated_blocks_and_data(blocks: np.ndarray) -> tuple[np.ndarray, list[bool]]:
    """
    Given an array of blocks, iterate over them and conjugate them if needed, and record the conjugation operations in a
    conjugation map. This function uses 0.5 as the threshold for conjugation instead of alpha0, to increase the noise
    of the conjugation map.
    :param blocks: the array of blocks
    :return: the mapped blocks, with their matching conjugation map
    """
    conjugation_map = []
    i = 0
    for block in blocks:
        if calc_bpcs_complexity_coefficient(block) >= 0.5:
            conjugation_map.append(False)
        else:
            conjugation_map.append(True)
            blocks[i] = conjugate(block)
        i += 1
    return blocks, conjugation_map


def embed_message_in_vessel(vessel_blocks: np.ndarray, alpha: float, message_blocks: np.ndarray,
                            message_bit_length: int, block_shape: tuple[int, int], check_capacity) -> np.ndarray:
    """
    Embeds an array of given message blocks and a given bit length into given vessel blocks.
    :param check_capacity: should we check the capacity of the vessel blocks before starting modify the vessel blocks?
    :param vessel_blocks: the blocks to embed the message into
    :param alpha: the minimum complexity coefficient threshold for each block, so that we know which bit plane blocks
    to modify in the vessel blocks.
    :param message_blocks: the message blocks to embed into the vessel blocks
    :param message_bit_length: the length of the message in bits
    :param block_shape: the shape for each block in the given array of message_blocks
    :return: the vessel blocks after embedding the message into them
    :raises BPCSError: if given an incorrect complexity coefficient threshold
    :raises BPCSCapacityError: if the vessel blocks don't have enough capacity to embed all the needed data
    """
    update_logger = multiprocessing.get_logger()
    update_logger.info("Starting embedding process...")

    if not 0 <= alpha <= 0.5:
        raise BPCSError("The minimum complexity coefficient must be between 0 and 0.5")

    update_logger.info("Counting accepted blocks number...")
    accepted_blocks_coords = collect_accepted_blocks(vessel_blocks, vessel_blocks.shape, block_shape, alpha)

    if check_capacity:
        update_logger.info("Checking image capacity...")
        if calculate_embedding_blocks_num(len(accepted_blocks_coords), block_shape, alpha,
                                          message_bit_length) <= len(accepted_blocks_coords):
            update_logger.info("Image has enough capacity of the embedding data!")
        else:
            update_logger.warning("Image does not have enough capacity of the embedding data!")
            raise BPCSCapacityError("Image does not have enough capacity of the embedding data")

    update_logger.info("Building initialization vector blocks...")
    iv_blocks = build_iv_blocks(len(accepted_blocks_coords), block_shape, alpha, message_bit_length)

    update_logger.info("Building conjugation map blocks...")
    message_blocks, conj_map = get_conjugated_blocks_and_data(message_blocks)

    conjugation_blocks = build_conjugation_blocks(conj_map, block_shape, alpha)

    embedding_blocks = np.concatenate([iv_blocks, conjugation_blocks, message_blocks])

    update_logger.info(f"Embedding {len(embedding_blocks)} blocks in image blocks...")

    embedding_block_index = 0
    for block_coords in accepted_blocks_coords:
        if embedding_block_index >= len(embedding_blocks):
            break

        vessel_blocks[tuple(block_coords)] = embedding_blocks[embedding_block_index]
        embedding_block_index += 1

        if embedding_block_index % 10000 == 0:
            update_logger.info(f"Block #{embedding_block_index} of {len(embedding_blocks)}")

    if embedding_block_index < len(embedding_blocks):
        raise BPCSCapacityError("Image does not have enough capacity of the embedding data!")

    update_logger.info("Finished embedding process!")
    return vessel_blocks



import multiprocessing
from typing import Generator, Any

from ..errors import BPCSError


def compute_segment_division_indices(total_length: int, segment_length: int) -> list[tuple[int, Any]]:
    """
    Divides a range of length total_length into separate segments of all length segment_length. Leaves out any segments
    that don't have length of segment_length (if total_length%segment_length != 0 then we have segments that aren't of
    length segment_length).
    :param total_length: the length to divide
    :param segment_length: the length of each segment
    :return: a list of the starting and ending indices of each range in the form list[n] = (start, end)
    """
    startings = range(0, total_length, segment_length)
    endings = []
    for left in startings:
        end = min(total_length, left + segment_length)
        if end % segment_length == 0:
            endings.append(end)
    return list(zip(startings, endings))


def compute_all_block_indices(image_shape: tuple[int, int, int, int],
                              block_shape: tuple[int, int]) -> Generator[list[int | slice], Any, Any]:
    """
    Computes every possible block location in an image of shape (x, y, channel, bit-plane index).
    :param image_shape: an array describing the image to compute all grid dimensions from
    :param block_shape: a tuple of two integers representing the size of each bit plane
    :return: A generator in which every item is in the form of:
     slice(x_start, x_end, None), slice(y_start, y_end, None), channel_num, bit_plane_index
    :raises BPCSError: if given an image object with an incorrect shape
    """
    update_logger = multiprocessing.get_logger()
    if len(image_shape) != 4:
        raise BPCSError(f"Image shape does not match: (width, height, channel number, bits per channel value). "
                        f"Given shape: {image_shape}")

    x_segments = compute_segment_division_indices(image_shape[0], block_shape[0])
    y_segments = compute_segment_division_indices(image_shape[1], block_shape[1])

    num_of_total_blocks = (image_shape[2] * image_shape[3] * len(x_segments) * len(y_segments))

    update_logger.info(f"Found {num_of_total_blocks} blocks of shape {block_shape} in the image.")

    i = 0
    for bit_index in reversed(range(image_shape[3])):  # each bit index from lsb -> msb
        for channel_index in range(image_shape[2]):  # each channel
            for (x_left, x_right) in x_segments:  # width
                for (y_left, y_right) in y_segments:  # height
                    i += 1
                    if i % 10000 == 0:
                        update_logger.info(f"Block #{i} of {num_of_total_blocks}")
                    yield [slice(x_left, x_right), slice(y_left, y_right)] + [channel_index, bit_index]



import math
from random import choices
from typing import Generator, Any
import numpy as np


def bits_to_blocks(bit_list: list[bool], block_shape: tuple[int, int]) -> np.ndarray:
    """
    Converts a list of bits into blocks of a given shape.
    :param bit_list: list of bits  (every bit is a boolean value)
    :param block_shape: the shape of every wanted bit-plane block
    :return: a list of blocks in which any block is in the given shape
    """
    # calculate the remainder (if there is one)
    area = block_shape[0] * block_shape[1]
    rem = (len(bit_list) % area)
    length_missing = area - rem if rem else 0

    # correct the remainder
    bit_list += choices([True, False], k=length_missing)
    bit_list = np.array(bit_list)
    num_of_blocks = math.ceil(len(bit_list) / area)

    return np.resize(bit_list, [num_of_blocks, block_shape[0], block_shape[1]])


def bytes_to_blocks(message: bytes, block_shape: tuple[int, int]) -> np.ndarray:
    """
    Converts a bytestring in an array of blocks of a given shape that represent the bytestring.
    :param message: the bytestring to be converted
    :param block_shape: the wanted block shape
    :return: the bytestring in the form of blocks
    """
    def bits(st: bytes) -> Generator[bool, Any, None]:
        bytes_gen = (b for b in st)
        for b in bytes_gen:
            for i in reversed(range(8)):
                yield bool((b >> i) & True)

    bits_list = list(bits(message))  # get the list of bits from the message
    return bits_to_blocks(bits_list, block_shape)  # return structured blocks of the bits


def blocks_to_bits(blocks: np.ndarray) -> list[bool]:
    """
    Given a list of blocks of bits, convert the blocks into a large list of bits
    :param blocks: a list of bit blocks
    :return: the list of bits that the list of blocks contained, all in one dimension
    """
    blocks = [np.array(block).reshape(-1) for block in blocks]
    return np.hstack(blocks).flatten().tolist()



import math
from random import choices
import numpy as np

from .core import calc_bpcs_complexity_coefficient


def get_prefix_length(block_area: int, min_alpha: float) -> int:
    """
    Calculates the length of the prefix for a given block area and minimum complexity coefficient.
    :param block_area: the total number of bits in a block
    :param min_alpha: the minimum complexity coefficient that the prefix should fulfill
    """
    # calibrated rho function
    return math.ceil(block_area * (1.4*min_alpha + 0.05))


def get_next_dynamically_prefixed_block(bits: list[bool], block_shape: tuple[int, int], min_alpha: float):
    """
    Builds a singular dynamically-prefixed block that contains a part of the given bits.
    :param bits: the bits
    :param block_shape: the shape of each bit plane block
    :param min_alpha: the minimum complexity coefficient that the prefix should fulfill
    :return: the built block, and the remaining bits
    """
    block_area = block_shape[0] * block_shape[1]
    prefix_length = get_prefix_length(block_area, min_alpha)
    data_length = block_area - prefix_length
    block_data, bits = bits[:data_length], bits[data_length:]

    if min_alpha == 0.0:
        return np.reshape(block_data, block_shape), bits

    while True:
        block = np.concatenate([choices([True, False], k=prefix_length), block_data])

        if len(block) < block_area:
            block = np.concatenate([block, choices([True, False], k=block_area - len(block))])

        block = np.reshape(block, block_shape)

        if calc_bpcs_complexity_coefficient(block) >= min_alpha:
            return block, bits


def bits_to_prefixed_blocks(bits: list[bool], block_shape: tuple[int, int], min_alpha: float) -> np.ndarray:
    """
    Construct prefixed blocks from data bits, where each block has a complexity coefficient of min_alpha or greater.
    :param bits: the array of data bits
    :param block_shape: the shape of each constructed block
    :param min_alpha: the minimum complexity coefficient of each block
    :return: the prefixed blocks constructed from the bits
    """
    blocks = []
    while len(bits) > 0:
        prefixed_block, bits = get_next_dynamically_prefixed_block(bits, block_shape, min_alpha)
        blocks.append(prefixed_block)

    return np.reshape(np.array(blocks), (len(blocks),) + block_shape)


def get_data_from_prefixed_blocks(blocks: np.ndarray, block_shape: tuple[int, int], alpha: float,
                                  data_bit_length: int) -> np.ndarray:
    """
    Parses the dynamically prefixed blocks, and extracts the data that they contain.
    :param blocks: the prefixed blocks we want to parse data from
    :param block_shape: the shape of each bit plane block given
    :param alpha: the minimum complexity coefficient that each block was prefixed to match
    :param data_bit_length: the length of the data we want to extract from the prefixed blocks in bits
    :return: an array of bools that represent the data we got from the prefixed blocks
    """
    data = np.array([], dtype=bool)
    block_area = block_shape[0] * block_shape[1]
    prefix_length = get_prefix_length(block_area, alpha)
    for block in blocks:
        block = np.reshape(block, (-1))
        data = np.append(data, block[prefix_length:])
    return data[:data_bit_length]



import numpy as np


def max_bpcs_complexity(row_num: int, col_num: int) -> int:
    """
    Calculate the maximum BPCS complexity of a bit plane given its size. Using the formula for a plane of size m, n:
    max_comp = m(n-1) + n(m-1)
    :param row_num: grids' number of columns
    :param col_num: grids' number of rows
    :return: the maximum BPCS complexity of the given bit plane
    """
    return ((col_num - 1) * row_num) + ((row_num - 1) * col_num)


def calculate_bit_changes(bits: list[bool]) -> int:
    """
    Calculate the number of bit changes in a given list of bits.
    :param bits: the given collection of bits
    :return: the number of bit changes
    """
    return sum([int(bits[i] ^ bits[i - 1]) for i in range(1, len(bits))])


def calc_bpcs_complexity_coefficient(bit_plane: np.ndarray) -> float:
    """
    Calculate the BPCS complexity coefficient of a given bit plane.
    :param bit_plane: the bit plane
    :return: the BPCS complexity coefficient of the given bitmap
    """
    max_complexity = max_bpcs_complexity(bit_plane.shape[0], bit_plane.shape[1])
    k = 0
    for row in bit_plane:
        k += calculate_bit_changes(row)
    for col in bit_plane.transpose():
        k += calculate_bit_changes(col)
    return (k * 1.0) / max_complexity


def checkerboard(h: int, w: int) -> np.ndarray:
    """
    Build a bit plane checkerboard of height h and width w
    :param h: wanted height
    :param w: wanted width
    :return: a checkerboard array of shape == [h,w]
    """
    re = np.r_[int(w / 2) * [False, True] + ([False] if w % 2 else [])]
    ro = True ^ re
    return np.vstack(int(h / 2) * (re, ro) + ((re,) if h % 2 else ()))


def conjugate(bit_plane: np.ndarray) -> np.ndarray:
    """
    Conjugates a given bit plane so that its complexity coefficient is 1-a where 'a' is the complexity coefficient of
    the given bit plane.
    :param bit_plane: the given bit plane
    :return: the conjugated bit plane
    """
    wc = checkerboard(bit_plane.shape[0], bit_plane.shape[1])  # white pixel at origin
    bc = True ^ wc  # black pixel at origin
    return np.array([[wc[i, j] if bit_plane[i, j] else bc[i, j] for j, cell in enumerate(row)]
                     for i, row in enumerate(bit_plane)])



from reedsolo import RSCodec, ReedSolomonError


def pad(content: bytes, ecc_block_size: int, ecc_symbol_num: int) -> bytes:
    """
    Pads content with Reed-Solomon Error Correction Codes.
    :param content: the content to be padded
    :param ecc_block_size: the rs block size
    :param ecc_symbol_num: the bytes per block that we want to dedicate for error correction
    :return: the padded content
    :raises ReedSolomonError: if the function was fed invalid parameters
    """
    if ecc_symbol_num == 0:
        return content

    if ecc_block_size > 255:
        # According to the reedsolo docs, if the block size is larger than 255, complexity and computational cost goes
        # up quadratically. So we limit the maximum block size to be 255.
        raise ReedSolomonError(f"Block size is too large: {ecc_block_size}")
    if ecc_symbol_num >= ecc_block_size:
        raise ReedSolomonError(f"ECC Symbol num is larger than the block size: {ecc_symbol_num} > {ecc_block_size}")
    rsc = RSCodec(ecc_symbol_num, nsize=ecc_block_size)
    return bytes(rsc.encode(content))


def unpad(padded: bytes, block_size: int, ecc_symbol_num: int) -> bytes:
    """
    Unpads padded content with Reed-Solomon Error Correction Codes.
    :param padded: the padded content
    :param block_size: the rs block size
    :param ecc_symbol_num: the bytes per block that we want to dedicate for error correction
    :return: the unpadded content
    :raises ReedSolomonError: if the function was fed invalid parameters, or the data had too many errors for the rs
    algorithm to fix.
    """
    if ecc_symbol_num == 0:
        return padded

    if block_size > 255:
        # According to the reedsolo docs, if the block size is larger than 255, complexity and computational cost goes
        # up quadratically. So we limit the maximum block size to be 255.
        raise ReedSolomonError(f"Block size is too large: {block_size}")
    if ecc_symbol_num >= block_size:
        raise ReedSolomonError(f"ECC Symbol num is larger than the block size: {ecc_symbol_num} > {block_size}")
    rsc = RSCodec(ecc_symbol_num, nsize=block_size)
    return rsc.decode(padded)[0]



import math
import numpy as np

from .core import calc_bpcs_complexity_coefficient
from .dimension_computing import compute_all_block_indices
from .dynamic_prefixing import get_prefix_length
from ..content_wrapper.wrapper import get_max_unwapped_length


def count_accepted_blocks(vessel_blocks: np.ndarray, image_shape: tuple[int, int, int, int],
                          block_shape: tuple[int, int], alpha: float) -> int:
    """
    Counts the number of accepted blocks. The threshold for being an accepted block is to have a complexity
    coefficient of alpha or greater.
    :param block_shape: the shape of each bit plane block
    :param image_shape: the shape of image
    :param vessel_blocks: the images' vessel blocks
    :param alpha: the minimum complexity coefficient threshold
    :return: the number of accepted blocks in the entire image
    """
    bit_plane_dims = compute_all_block_indices(image_shape, block_shape)
    noise_blocks_num = 0
    for bit_plane in bit_plane_dims:
        block = vessel_blocks[tuple(bit_plane)]  # get the current block to handle
        if calc_bpcs_complexity_coefficient(block) >= alpha:  # check if the block is valid for embedding data
            # if the block as a complexity coefficient that is sufficient for embedding data add 1 to the counter
            noise_blocks_num += 1
    return noise_blocks_num


def collect_accepted_blocks(vessel_blocks: np.ndarray, image_shape: tuple[int, int, int, int],
                            block_shape: tuple[int, int], alpha: float):
    bit_plane_dims = compute_all_block_indices(image_shape, block_shape)

    accepted_blocks = []
    for bit_plane in bit_plane_dims:
        block = vessel_blocks[tuple(bit_plane)]  # get the current block to handle
        if calc_bpcs_complexity_coefficient(block) >= alpha:
            accepted_blocks.append(tuple(bit_plane))
    return accepted_blocks


def calculate_embedding_blocks_num(accepted_blocks_num: int, block_shape: tuple[int, int], alpha: float,
                                   message_bit_length: int) -> int:
    """
    Calculates the number of blocks we need to modify to embed all the given data.
    :param accepted_blocks_num: the number of accepted blocks in the image
    :param block_shape: the shape of each bit plane block
    :param alpha: the minimum complexity coefficient threshold
    :param message_bit_length: the bit length of the message
    :return: the number of total block length of the embedding payload
    """
    block_area = block_shape[0] * block_shape[1]
    bits_per_prefixed_block = block_area - get_prefix_length(block_area, alpha)

    iv_bit_length = len(bin(accepted_blocks_num)[2:]) + len(bin(block_area)[2:])
    conjugation_map_bit_length = math.ceil(message_bit_length / block_area)

    iv_block_length = math.ceil(iv_bit_length / bits_per_prefixed_block)
    conjugation_map_block_length = math.ceil(conjugation_map_bit_length / bits_per_prefixed_block)
    message_block_length = math.ceil(message_bit_length / block_area)

    return iv_block_length + conjugation_map_block_length + message_block_length


def calculate_maximum_capacity(vessel_blocks: np.ndarray, image_shape: tuple[int, int, int, int],
                               ecc_block_size: int, ecc_symbol_num: int, alpha: float) -> int:
    block_shape_length = 8
    accepted_blocks_num = count_accepted_blocks(vessel_blocks, image_shape,
                                                (block_shape_length, block_shape_length), alpha)
    bits_per_prefixed_block = block_shape_length ** 2 - get_prefix_length(block_shape_length ** 2, alpha)
    iv_bit_length = math.ceil(math.log2(accepted_blocks_num)) + math.ceil(math.log2(block_shape_length ** 2))

    max_bit_embedding_input_length = math.floor(
        ((block_shape_length ** 2) * ((bits_per_prefixed_block * (accepted_blocks_num - 3)) - (iv_bit_length + 1)))
        /
        (bits_per_prefixed_block + 1)
    )

    return get_max_unwapped_length(max_bit_embedding_input_length, ecc_block_size, ecc_symbol_num)



import random
import numpy


def get_shuffled_perm(length: int, seed: bytes):
    perm = list(numpy.arange(length) + 1)
    random.seed(seed)
    random.shuffle(perm)
    return perm


def shuffle_bytes(raw_bytes: bytes, seed: bytes):
    # Shuffle the list ls using the seed `seed`
    raw_bytes = bytearray(raw_bytes)
    random.seed(seed)
    random.shuffle(raw_bytes)
    return bytes(raw_bytes)


def unshuffle_bytes(shuffled_bytes: bytes, seed: bytes):
    shuffled_perm = get_shuffled_perm(len(shuffled_bytes), seed)

    zipped_ls = list(zip(shuffled_bytes, shuffled_perm))
    zipped_ls.sort(key=lambda x: x[1])
    unshuffled = bytearray()
    [unshuffled.append(a) for (a, b) in zipped_ls]
    return bytes(unshuffled)



from .bpcs.bpcs_image import BPCSImage
from .bpcs.embed import get_message_blocks_from_bytes
from .bpcs.capacity import calculate_maximum_capacity
from .image_utils import image_to_bytes
from .lsb.lsb_image import LSBImage
from .content_wrapper.wrapper import wrap_bpcs, get_bpcs_token_info, wrap_lsb, get_lsb_token_info, \
     unwrap, get_max_unwapped_length
from .steganalysis.bit_plane_slicing import slice_rgb_bit_planes
from .steganalysis.get_diff import show_diff


def bpcs_embed(source_image_bytes: bytes, message: bytes, key: str, ecc_block_size: int = 255,
               ecc_symbol_num: int = 16, alpha: float = 0.3, check_capacity=True) -> tuple[bytes, bytes]:
    """
    Embeds message into the image at source_image_path, affecting blocks that have a
    complexity coefficient of alpha or greater, then saves the resulting image to output_file_path.
    """
    message, token = wrap_bpcs(message, key.encode(), ecc_block_size, ecc_symbol_num, alpha)
    img = BPCSImage(source_image_bytes, as_cgc=True)
    message_blocks, message_bit_length = get_message_blocks_from_bytes(message)
    arr = img.embed(message_blocks, message_bit_length, alpha, check_capacity)
    new_image_bytes = img.export(arr)
    return new_image_bytes, token


def bpcs_extract(source_image_bytes: bytes, token: bytes) -> bytes:
    """
    Extracts data from the image at source_image_path, and returns the data.
    """
    ((ecc_block_size, ecc_symbol_num), (verification_tag, nonce, update_header, key),
     seed, min_alpha) = get_bpcs_token_info(token)
    img = BPCSImage(source_image_bytes, as_cgc=True)
    wrapped = img.extract(min_alpha)
    return unwrap(wrapped, ecc_block_size, ecc_symbol_num, verification_tag, nonce, update_header, seed, key)


def bpcs_calculate_max_capacity(source_image_bytes: bytes, ecc_block_size: int = 255,
                                ecc_symbol_num: int = 16, alpha: float = 0.3) -> int:
    img = BPCSImage(source_image_bytes, as_cgc=True)
    image_shape = img.pixels.shape
    max_message_byte_length = calculate_maximum_capacity(img.pixels, image_shape, ecc_block_size, ecc_symbol_num, alpha)
    return max_message_byte_length


def lsb_embed(source_image_bytes: bytes, message: bytes, key: str, ecc_block_size: int = 255,
              ecc_symbol_num: int = 16, num_of_sacrificed_bits: int = 2, check_capacity=True) -> tuple[bytes, bytes]:
    message, token = wrap_lsb(message, key.encode(), ecc_block_size, ecc_symbol_num, num_of_sacrificed_bits)
    img = LSBImage(source_image_bytes, num_of_sacrificed_bits)
    new_image = img.embed(message, check_capacity)
    return image_to_bytes(new_image), token


def lsb_extract(source_image_bytes: bytes, token: bytes) -> bytes:
    ((ecc_block_size, ecc_symbol_num), (verification_tag, nonce, update_header, key),
     seed, num_of_sacrificed_bits) = get_lsb_token_info(token)
    img = LSBImage(source_image_bytes, num_of_sacrificed_bits)
    wrapped = img.extract()
    return unwrap(wrapped, ecc_block_size, ecc_symbol_num, verification_tag, nonce, update_header, seed, key)


def lsb_calculate_max_capacity(source_image_bytes: bytes, ecc_block_size: int = 255, ecc_symbol_num: int = 16,
                               num_of_sacrificed_bits: int = 2) -> int:
    img = LSBImage(source_image_bytes, num_of_sacrificed_bits)
    total_available_bits = img.image.width * img.image.height * len(img.image.getbands()) * num_of_sacrificed_bits
    iv_bit_len = img.iv_bit_len
    max_bit_embedding_input_length = total_available_bits - iv_bit_len
    return get_max_unwapped_length(max_bit_embedding_input_length, ecc_block_size, ecc_symbol_num)


def get_image_diffences(img1_bytes: bytes, img2_bytes: bytes, exact_diff: bool) -> tuple[tuple[int, int, int], bytes]:
    return show_diff(img1_bytes, img2_bytes, exact_diff)


def slice_image_bitplanes(image_bytes: bytes) -> list[tuple[str, bytes]]:
    bitplanes = []
    for name, bitplane_bytes in slice_rgb_bit_planes(image_bytes):
        bitplanes.append((name, bitplane_bytes))
    return bitplanes



import math
import numpy as np

from .dynamic_prefixing import get_prefix_length, bits_to_prefixed_blocks, \
    get_data_from_prefixed_blocks


def bit_array_to_int(bit_array: np.ndarray) -> int:
    """
    Given a bit array, convert it into a decimal integer and return it.
    :param bit_array: an array of bits (bools)
    :return: the decimal integer representation of the bit array value
    """
    return int("".join([str(int(i)) for i in bit_array.tolist()]), 2)


def build_message_length_iv(number_of_accepted_blocks: int, block_shape: tuple[int, int],
                            message_bit_length: int) -> np.ndarray:
    """
    Builds an initialization vector that contains the message block length.
    :param number_of_accepted_blocks: the number of accepted blocks
    :param block_shape: the shape of each bit plane block in the image
    :param message_bit_length: the length of the message in bits
    :return: the initialization vector that contains the message block length, sized to a static length for that image
    with a specific block shape and complexity coefficient.
    """
    message_block_length = math.ceil(message_bit_length / (block_shape[0] * block_shape[1]))
    iv = str.zfill(bin(message_block_length)[2:], len(bin(number_of_accepted_blocks)[2:]))
    return np.array([bool(int(i)) for i in iv])


def build_message_remnant_iv(block_shape: tuple[int, int], message_bit_length: int) -> np.ndarray:
    """
    Builds an initialization vector that contains the number of bits in the last message block that belong to the
    message and aren't just filler bits.
    :param block_shape: the shape of each bit plane block in the image
    :param message_bit_length: the length of the message in bits
    :return: initialization vector that contains the number of bits in the last message block that belong to the
    message and aren't just filler bits, sized to a constant length of the specific block shape.
    """
    block_area = block_shape[0] * block_shape[1]
    message_remnant_bits_num = message_bit_length % block_area
    if message_remnant_bits_num == 0:
        message_remnant_bits_num = block_area
    iv = str.zfill(str(bin(message_remnant_bits_num)[2:]), len(bin(block_area)[2:]))
    return np.array([bool(int(i)) for i in iv])


def build_iv_blocks(number_of_accepted_blocks: int, block_shape: tuple[int, int],
                    alpha: float, message_bit_length: int) -> np.ndarray:
    """
    Builds the initialization vector blocks given the needed parameters. Each initialization vector has a constant
    length that depends on the number of accepted blocks, the block shape, and the minimum complexity coefficient.
    :param number_of_accepted_blocks: the number of bit plane blocks in the entire image that have a complexity
    coefficient of alpha or greater
    :param block_shape: the shape of each block in the image, and the wanted shape of each iv block
    :param alpha: the minimum complexity coefficient that each iv block needs to be prefixed to match
    :param message_bit_length: the length of the message in bits
    :return: an array of the initialization vector blocks
    """
    message_length_iv = build_message_length_iv(number_of_accepted_blocks, block_shape, message_bit_length)
    message_remnant_iv = build_message_remnant_iv(block_shape, message_bit_length)
    iv_bits = np.append(message_length_iv, message_remnant_iv)
    return bits_to_prefixed_blocks(iv_bits.tolist(), block_shape, min_alpha=alpha)


def build_conjugation_blocks(conjugation_map: list[bool], block_shape: tuple[int, int], alpha: float) -> np.ndarray:
    """
    Builds the conjugation blocks.
    :param conjugation_map: a list of bits that represent the conjugation map of the embedded message
    :param block_shape: the shape of each bit plane block in the image
    :param alpha: the minimum complexity coefficient that each iv block needs to be prefixed to match
    :return: the conjugation map blocks, prefixed to have a complexity coefficient of alpha or greater
    """
    return bits_to_prefixed_blocks(conjugation_map, block_shape, alpha)


def slice_iv_from_accepted_blocks(blocks: np.ndarray, block_shape: tuple[int, int],
                                  alpha: float) -> tuple[tuple[int, int], np.ndarray]:
    """
    Finds, separates, and extracts data from the initialization vector blocks in an array of image's bit plane blocks,
    using the constant length of the iv for every image and its parameters. This function assumes the relevant blocks
    begin at index = 0 of the given array of blocks.
    :param blocks: the images' bit plane blocks that have a complexity coefficient of alpha or more
    :param block_shape: the shape of each bit plane block in the image
    :param alpha: the minimum complexity coefficient of each block that was given
    :return: the data from the initialization vector blocks, and the remaining block in the form:
    (message_block_length, message_remnant_bits), remaining blocks
    """
    # calculate what's the length of the iv in blocks, using the constant length for each iv part
    block_area = block_shape[0] * block_shape[1]
    iv_bit_length = len(bin(len(blocks))[2:]) + len(bin(block_area)[2:])
    available_bits_per_block = block_area - get_prefix_length(block_area, alpha)
    iv_block_length = math.ceil(iv_bit_length / available_bits_per_block)

    # separate the iv blocks from the rest of the blocks, and extract the data from the iv blocks
    iv_blocks, remaining_blocks = blocks[:iv_block_length], blocks[iv_block_length:]
    iv_bits = get_data_from_prefixed_blocks(iv_blocks, block_shape, alpha, iv_bit_length)
    message_length_iv_len = len(bin(len(blocks))[2:])
    message_length_iv, message_remnant_iv = iv_bits[:message_length_iv_len], iv_bits[message_length_iv_len:]
    return (bit_array_to_int(message_length_iv), bit_array_to_int(message_remnant_iv)), remaining_blocks


def slice_conj_blocks_from_accepted_blocks(blocks: np.ndarray, block_shape: tuple[int, int], alpha: float,
                                           embedded_message_blocks_num: int) -> tuple[list[bool], np.ndarray]:
    """
    Extracts the conjugation map from the images' accepted blocks.
    :param blocks: the images' bit plane blocks that have a complexity coefficient of alpha or more
    :param alpha: the minimum complexity coefficient of each block that was given
    :param block_shape: the shape of each bit plane block in the image
    :param embedded_message_blocks_num: the number of embedded message blocks in the image, this is also the length of
    the conjugation map in bits.
    :return: the conjugation map blocks, and the remaining blocks in the form: (map, remaining blocks)
    """
    conj_map_bit_length = embedded_message_blocks_num
    block_area = block_shape[0] * block_shape[1]

    conj_map_block_length = math.ceil(conj_map_bit_length / (block_area - get_prefix_length(block_area, alpha)))

    conj_blocks, remaining = blocks[:conj_map_block_length], blocks[conj_map_block_length:]

    return get_data_from_prefixed_blocks(conj_blocks, block_shape, alpha, conj_map_bit_length).tolist(), remaining



class ContentWrapperError(Exception):
    pass


class TokenError(ContentWrapperError):
    pass


class SteganographyError(Exception):
    pass


class BPCSError(SteganographyError):
    """
    The parent error of all BPCS related errors.
    """
    pass


class BPCSCapacityError(BPCSError):
    """
    Errors that deal with the capacity of an image.
    """
    pass


class BPCSEmbedError(BPCSError):
    """
    Errors that deal with the embedding process of data in an image.
    """
    pass


class BPCSExtractError(BPCSError):
    """
    Errors that deal with the extracting process of data in an image.
    """
    pass


class LSBError(SteganographyError):
    """
    The parent error of all LSB related errors.
    """
    pass


class LSBCapacityError(LSBError):
    """
    Errors that deal with the capacity of an image.
    """
    pass


class LSBEmbedError(LSBError):
    """
    Errors that deal with the embedding process of data in an image.
    """
    pass


class LSBExtractError(LSBError):
    """
    Errors that deal with the extracting process of data in an image.
    """
    pass



import numpy as np


def bits_to_bytes(bits: list[bool]) -> bytes:
    """
    Converts a list of bits into bytes.
    :param bits: the given list of bits
    :return: a bytes type value, representing the value of the list of bits
    """
    spare_bits_length = len(bits) % 8  # calculate the length of any spare bits

    # since the message was initially read by the byte, any spares must have been added to create a block
    bits = bits[:len(bits) - spare_bits_length]  # get rid of any spare bits

    bytes_number = int(len(bits) / 8)
    message_bytes = np.resize(np.array(bits), [bytes_number, 8])

    def byte_to_decimal_int(byte: np.ndarray) -> int:
        return int('0b' + ''.join(str(int(x)) for x in byte.tolist()), 2)

    def decimal_int_to_bytes(byte) -> bytes:
        return byte_to_decimal_int(byte).to_bytes()

    return b''.join([decimal_int_to_bytes(byte) for byte in message_bytes])


def bit_list_to_int(bitlist: list[bool]) -> int:
    return int("".join((str(int(i)) for i in bitlist)), 2)


def bytes_to_bit_list(st: bytes, length: int = None) -> list[bool]:
    bitlist = []
    bytes_gen = (b for b in st)
    for b in bytes_gen:
        for i in reversed(range(8)):
            bitlist.append(bool((b >> i) & True))
    if length is not None:
        bitlist = (max(length - len(bitlist), 0) * [False]) + bitlist
    return bitlist


def bitlist_str_to_list(bitlist_str: str, length: int):
    bitlist = list(bool(int(i)) for i in bitlist_str)
    bitlist = (max(length - len(bitlist), 0) * [False]) + bitlist
    return bitlist



import multiprocessing
from math import log2, ceil
from typing import Generator, Any

from ..bit_operations_utils import bits_to_bytes, bit_list_to_int, bytes_to_bit_list, \
    bitlist_str_to_list
from ..image_utils import open_image_from_bytes
from ..errors import LSBError, LSBCapacityError

# Mask used to put one (ex:1->00000001, 2->00000010) associated with OR bitwise
TRUE_BIT_MASK_VALUES = [1, 2, 4, 8, 16, 32, 64, 128]

# Mask used to put zero (ex:254->11111110, 253->11111101) associated with AND bitwise
FALSE_BIT_MASK_VALUES = [254, 253, 251, 247, 239, 223, 191, 127]


def construct_iv(data_bit_length: int, iv_bit_length: int) -> list[bool]:
    iv = bitlist_str_to_list(bin(data_bit_length)[2:], iv_bit_length)
    iv = iv[:iv_bit_length]
    return iv


class LSBImage:
    def __init__(self, image_bytes, sacrificed_bits=2):
        update_logger = multiprocessing.get_logger()
        update_logger.info("Loading image...")
        self.image = open_image_from_bytes(image_bytes)
        self.iv_bit_len = ceil(log2(self.image.width*self.image.height*len(self.image.getbands())*sacrificed_bits))
        self.max_bits_per_byte = sacrificed_bits

        if not 0 < self.max_bits_per_byte <= 8:
            raise LSBError("The number of bits dedicated to storing data per byte must be 0 < num <= 8")

        self.one_mask_values = TRUE_BIT_MASK_VALUES[:sacrificed_bits]
        self.one_mask_max = self.one_mask_values[-1]
        self.one_mask = self.one_mask_values.pop(0)

        self.zero_max_values = FALSE_BIT_MASK_VALUES[:sacrificed_bits]
        self.zero_mask_max = self.zero_max_values[-1]
        self.zero_mask = self.zero_max_values.pop(0)

        self.cursor_width = 0  # Current width position
        self.cursor_height = 0  # Current height position
        self.cursor_channel = 0  # Current channel position
        update_logger.info("Loaded Image!")

    def increment_cursor(self):
        if self.cursor_width < self.image.width - 1:
            self.cursor_width += 1
            return
        self.cursor_width = 0

        if self.cursor_height < self.image.height - 1:
            self.cursor_height += 1
            return
        self.cursor_height = 0

        if self.cursor_channel < len(self.image.getbands()) - 1:
            self.cursor_channel += 1
            return
        self.cursor_channel = 0

        if self.one_mask < self.one_mask_max:
            self.one_mask = self.one_mask_values.pop(0)
            self.zero_mask = self.zero_max_values.pop(0)
            return

        raise LSBCapacityError("No available slot remaining (image filled)")

    def put_binary_value(self, bits: list[bool] | Generator[bool, Any, None]):
        for bit in bits:
            pixel = list(self.image.getpixel((self.cursor_width, self.cursor_height))) # type: ignore
            byte_value = int(pixel[self.cursor_channel])
            if bit:
                pixel[self.cursor_channel] = byte_value | self.one_mask  # bitwise OR with one_mask
            else:
                pixel[self.cursor_channel] = byte_value & self.zero_mask  # bitwise AND with one_mask

            self.image.putpixel((self.cursor_width, self.cursor_height), tuple(pixel))
            self.increment_cursor()

    def read_bit(self) -> bool:
        value_byte = self.image.getpixel((self.cursor_width, self.cursor_height))[self.cursor_channel] # type: ignore
        value_byte = int(value_byte) & self.one_mask
        self.increment_cursor()
        if value_byte > 0:
            return True
        else:
            return False

    def read_bits(self, num_of_bits: int) -> list[bool]:
        bits = []
        for i in range(num_of_bits):
            bits.append(self.read_bit())
        return bits

    def read_byte(self) -> list[bool]:
        return self.read_bits(8)

    def embed(self, data: bytes, check_capacity: bool):
        update_logger = multiprocessing.get_logger()
        update_logger.info("Starting embedding process...")

        data_byte_length = len(data)

        if check_capacity:
            update_logger.info("Checking if image has enough capacity to contain the sent data...")
            if not self.check_capacity(data_byte_length*8):
                update_logger.info("Image does NOT have enough capacity!")
                raise LSBCapacityError("Carrier image not big enough to contain all the data to embed.")
            update_logger.info("Image has enough capacity!")

        iv_bits = construct_iv(data_byte_length*8, self.iv_bit_len)
        data_bits = bytes_to_bit_list(data)

        self.put_binary_value(iv_bits)
        self.put_binary_value(data_bits)

        update_logger.info("Finished embedding process!")

        return self.image

    def extract(self):
        update_logger = multiprocessing.get_logger()
        update_logger.info("Starting reading process...")

        iv = self.read_bits(self.iv_bit_len)
        data_length = bit_list_to_int(iv)
        output = self.read_bits(data_length)

        update_logger.info("Reading process finished!")
        return bits_to_bytes(output)

    def check_capacity(self, data_bit_length: int) -> bool:
        update_logger = multiprocessing.get_logger()
        update_logger.info("Calculating if sent data will fit into the image...")
        return ((self.image.width * self.image.height * len(self.image.getbands()) * self.max_bits_per_byte)
                >= data_bit_length + self.iv_bit_len)



from .engine import bpcs_embed, bpcs_extract, bpcs_calculate_max_capacity, lsb_embed, lsb_extract, \
                    lsb_calculate_max_capacity, get_image_diffences, slice_image_bitplanes
from .errors import ContentWrapperError, TokenError, SteganographyError, BPCSError, BPCSCapacityError, \
                    BPCSEmbedError, BPCSExtractError, LSBError, LSBCapacityError, LSBEmbedError, \
                    LSBExtractError


import hashlib
from Crypto.Cipher import AES
from Crypto.Random import get_random_bytes

from ..errors import ContentWrapperError


def encrypt(plaintext: bytes, key: bytes) -> tuple[bytes, bytes, bytes, bytes, bytes]:
    """
    Encrypts plaintext using AES-256 GCM algorithm.
    :param plaintext: the plaintext to be encrypted
    :param key: the encryption key
    :return: the encrypted ciphertext, with the relevant info in the form of: ciphertext, tag, nonce, header, key
    """
    key_hash = hashlib.sha256(key).digest()
    update_header = get_random_bytes(8)
    cipher = AES.new(key_hash, AES.MODE_GCM, mac_len=16)
    cipher.update(update_header)
    ciphertext, verification_tag = cipher.encrypt_and_digest(plaintext)
    return ciphertext, verification_tag, cipher.nonce, update_header, key


def decrypt(ciphertext: bytes, key: bytes, tag: bytes, nonce: bytes, update_header: bytes) -> bytes:
    """
    Decrypts ciphertext using AES-256 GCM algorithm.
    :param ciphertext: the ciphertext to be decrypted
    :param key: the encryption key
    :param tag: the GCM tag
    :param nonce: the GCM nonce
    :param update_header: the GCM header
    :return: the decrypted plaintext
    :raises ContentWrapperError: if the decryption process fails
    """
    key_hash = hashlib.sha256(key).digest()
    cipher = AES.new(key_hash, AES.MODE_GCM, mac_len=16, nonce=nonce)
    try:
        cipher.update(update_header)
        plaintext = cipher.decrypt_and_verify(ciphertext, tag)
    except (ValueError, KeyError) as e:
        # something went wrong with the decryption process (tampered or incorrect data)
        raise ContentWrapperError(f"Data decryption was unsuccessful. {e}")
    return plaintext
